Require Import Equations.Equations.

Inductive vect (A : Type) : nat -> Type :=
| vnil : vect A O
| vcons : forall (n : nat), A -> vect A n
     -> vect A (S n).

Arguments vnil {_}.
Arguments vcons {_ _} _ _.

(* Basic usage of Equations. *)

Equations tail {A : Type}
  {n : nat} (v : vect A (S n)) : vect A n :=
tail (vcons _ v') := v'.

Check @tail.
Check tail_equation_1.

Print tail_obligation_1.

Definition tail' {A : Type}
  {n : nat} (v : vect A (S n)) : vect A n :=
match v with
| vcons _ v' => v'
end.

Print tail'.

Equations dtail {A : Type}
  {n : nat} (v : vect A (S (S n))) : vect A n :=
dtail (vcons _ (vcons _ v')) := v'.

Fail Definition dtail' {A : Type}
  {n : nat} (v : vect A (S (S n))) : vect A n :=
match v with
| vcons _ (vcons _ v') => v'
end.


(* "with" clause *)

Equations unzip {A B : Type} {n : nat}
  (v : vect (A * B) n) :
  vect A n * vect B n :=
unzip vnil := (vnil, vnil) ;
unzip (vcons (pair a b) v) <= unzip v => {
  | pair v w => (vcons a v, vcons b w)
}.

Check @unzip.
Check unzip_equation_1.
Check unzip_equation_2.
Check unzip_helper_1_equation_1.

Equations nat_eqdec (n m : nat) :
  { n = m } + { n <> m } :=
nat_eqdec O O := left eq_refl ;
nat_eqdec (S n) (S m) <= nat_eqdec n m => {
    (* { n = n } + { n <> n } *)
  | left eq_refl := left eq_refl ;
    (* n <> m -> S n <> S m *)
  | right _ := right _
} ;
nat_eqdec _ _ := right _.



(* Functional elimination. *)

Require Import List.
Equations filter {A : Type} (P : A -> bool)
  (l : list A) : list A :=
filter P nil := nil;
filter P (cons a l) <= P a => {
  | true := cons a (filter P l);
  | false := filter P l
}.

Lemma filter_correct {A : Type} (P : A -> bool)
  (l : list A) :
  Forall (fun a => P a = true) (filter P l).
Proof.
  funelim (filter P l).
  - apply Forall_nil.
  - (* P a = true *)
    apply Forall_cons.
    + assumption.
    + assumption.
  - (* P a = false *)
    assumption.
Qed.
